package F5::CfgConvert;

use warnings;
use strict;

use YAML::XS;
use File::Basename;
use Bit::Vector;
use Fcntl 'SEEK_SET';
use MIME::Base64;
use File::Slurp;
use Math::BigInt;

use F5::Deploy;
use F5::IpAddr;
use Google::ProtocolBuffers;

our $DCC_TABLE_CFG_FILENAME = F5::Deploy::map_path('/etc/ts/common/dcc_tables_cfg.yaml');
our $BD_PROTO_PATH = F5::Deploy::map_path('/usr/share/ts/share/proto/');
our $BD_CONFIG_PROTO_FILENAME = 'bd_config.proto';
our $dcc_table_config;
our $MSG_SIGNATURE_DEFAULT = 98716425;
our $PROTOBUF_DCC_TABLE_ID = 203;
our $PREFIX_PACK_STRING = 'i i i';
our $PREFIX_BYTE_SIZE = 12;
our $PROTOBUF_MAX_BULK_SIZE = 1000;

my %types = (
             'binary' => {
                          pack_as => 'a',
                          field_length => 1,
                          fixup_func => sub {
                              MIME::Base64::decode_base64(shift); # because PrepareFiles does base64 encode for binary* type
                          },
                          use_fixup_func_for_protobuf => 1,
                         },
             'bitmap64' => {
                            pack_as => 'L L',
                            field_length => 8,
                            fixup_func => sub {
                                my $vec = Bit::Vector->new_Bin(64, shift);
                                return ($vec->Chunk_Read(32, 0), $vec->Chunk_Read(32, 32));
                            },
                           },
             'bitmap' => {
                            pack_as => 'L L',
                            field_length => 8,
                            protobuf_fixup_func => sub {
                                return Math::BigInt->new('0b'.shift);
                            },
                         },
             'byte' => {
                        pack_as => 'C',
                        field_length => 1,
                       },
             'byte256' => {
                           pack_as => 'c256',
                           field_length => 256,
                           fixup_func => sub {
                               return split(//, shift);
                            },
                       },
             'char' => {
                        pack_as => 'a',
                        field_length => 1,
                        fixup_func => sub {
                            my $string = shift;
                            $string =~ tr/\002\003\004/,\r\n/;
                            return $string;
                        },
                        use_fixup_func_for_protobuf => 1,
                       },
             'crc' => {
                       pack_as => 'L L ',
                       fixup_func => sub {
                           my $vec = Bit::Vector->new_Dec(64, shift);
                           return ($vec->Chunk_Read(32, 0), $vec->Chunk_Read(32, 32));
                       },
                       field_length => 8,
                      },
             'crc16' => {
                         pack_as => 'S',
                         field_length => 2,
                        },
             'crc32' => {
                         pack_as => 'L',
                         field_length => 4,
                        },
             'double' => {
                          pack_as => 'd', # todo:rss: is this long double
                          field_length => 8,
                         },
             'int' => {
                       pack_as => 'i',
                       field_length => 4,
                       empty_value => 0,
                      },
             'ip' => {
                      pack_as => 'a16',
                      field_length => 16,
                      fixup_func => sub {
                          return F5::IpAddr::ip_pton(shift);
                      }
                     },
             'long' => {
                        pack_as => 'L',
                        field_length => 4,
                        empty_value => 0,
                       },
             'mac' => {
                       pack_as => 'H12',
                       field_length => 6,
                       fixup_func => sub {
                           my $val = shift;
                           $val =~ s/://g;
                           return $val;
                       },
                      },
             'short' => {
                         pack_as => 's',
                         field_length => 2,
                        },
             'unsigned_char' => {
                                 pack_as => 'C',
                                 field_length => 1,
                                },
             'unsigned_int' => {
                                pack_as => 'I',
                                field_length => 4,
                                empty_value => 0,
                               },
             'unsigned_short' => {
                                  pack_as => 'S',
                                  field_length => 2,
                                  empty_value => 0,
                                 },
            );

prepare_protobuf_modules();

# Bd::Config::ConfigAction::* constants are loaded by
#     'Google::ProtocolBuffers->parse', below, in 'sub prepare_protobuf_modules',
# this section has to be after the call to 'prepare_protobuf_modules' !!!
my %PROTOBUF_OPERATION_MAP = (
                              '1' => Bd::Config::ConfigAction::ADD(),
                              '+1' => Bd::Config::ConfigAction::ADD(),
                              '-1' => Bd::Config::ConfigAction::DELETE(),
                              '0' => Bd::Config::ConfigAction::UPDATE(),
                             );

#----------------------------------------------------------------------

sub prepare_protobuf_modules {
    my $protobuf_cont = read_file($BD_PROTO_PATH.$BD_CONFIG_PROTO_FILENAME);
    Google::ProtocolBuffers->parse($protobuf_cont, {create_accessors => 1, include_dir => $BD_PROTO_PATH});
}

#----------------------------------------------------------------------

sub convert_table {
    my(%params) = @_;
    my $from_file = $params{from_file};

    my $dcc_table_config = get_dcc_table_config();

    my $from_file_cfg = basename($from_file);
    $from_file_cfg =~ s/\.cfg$//;
    $from_file_cfg =~ s/__account_(\d+)$//;
    $from_file_cfg =~ s/^.*__//;
    my($table_def) = grep { $from_file_cfg eq $_->{output_file} } @$dcc_table_config;

    return unless $table_def;

    if ($table_def->{protobuf_class_enum}) {
        return protobuf_convert(%params,
                                table_def => $table_def,
                               );
    }
    else {
        return binary_convert(%params,
                              table_def => $table_def,
                             );
    }
}

#----------------------------------------------------------------------

sub protobuf_convert {
    my(%params) = @_;

    my $table_def = $params{table_def};
    my $from_file = $params{from_file};
    my $protobuf_message_max_size = $params{protobuf_message_max_size};
    my $cfg_logger = $params{cfg_logger};

    if (! $table_def->{protobuf_convert_cfg}) {
        $table_def->{protobuf_convert_cfg} = {
                                              protobuf_class_enum => $table_def->{protobuf_class_enum},
                                              protobuf_entry_name => lc $table_def->{protobuf_class_enum},
                                             };
        my $protobuf_class = $table_def->{protobuf_convert_cfg}->{protobuf_entry_name};
        $protobuf_class =~ s/^(\w)/\u$1/;
        $protobuf_class =~ s/_(\w)/\u$1/g;
        $table_def->{protobuf_convert_cfg}->{message_protobuf_class_name} = 'Bd::Config::'. $protobuf_class;
        for my $orig_field_def (@{$table_def->{fields}}) {
            my $data_type = $orig_field_def->{data_type};
            $data_type =~ s/(\d+)$//;

            my $field_def = { %$orig_field_def,
                              %{$types{$data_type}},
                            };
            if ($field_def->{use_fixup_func_for_protobuf}) {
                $field_def->{protobuf_fixup_func} = $field_def->{fixup_func};
            }
            push(@{$table_def->{protobuf_convert_cfg}->{fields}}, $field_def);
        }
    }

    my $protobuf_convert_cfg = $table_def->{protobuf_convert_cfg};

    my $bd_conf_elem_from_prev_iteration;
    $cfg_logger->trace("Converting BD config from $from_file") if $cfg_logger;
    open(my $inf, '<', $from_file);
    my $iterator = sub {
        my $last_iteration = 1;
        return ['', $last_iteration] unless $inf;

        my $operation_count;

        my @config_elements;
        my $current_message_size = 0;
        if ($bd_conf_elem_from_prev_iteration) {
            push @config_elements, $bd_conf_elem_from_prev_iteration;
            $operation_count->{$bd_conf_elem_from_prev_iteration->{config_action}}++;
            $current_message_size = length($bd_conf_elem_from_prev_iteration) + 4;
            $bd_conf_elem_from_prev_iteration = undef;
        }

        while (my $line = <$inf>) {
            chomp($line);
            $cfg_logger->trace("Cfg line = $line") if $cfg_logger;
            my($rec_operation, @field_vals) = split(/,/, $line, -1);
            my $operation = $PROTOBUF_OPERATION_MAP{ $rec_operation };
            $operation ||= Bd::Config::ConfigAction::ADD();

            my $skip_line = 0;
            my $rec = {};
            for my $i (0..$#field_vals) {
                my $field_value = $field_vals[$i];
                my $field_def = $protobuf_convert_cfg->{fields}->[$i];

                if ($field_def->{protobuf_fixup_func}) {
                    my $rv;
                    eval {
                        $rv = $field_def->{protobuf_fixup_func}->($field_value);
                    };
                    if (my $err = $@) {
                        $cfg_logger->error("Error encountered while handling config line = $line, element = $i ($err)") if $cfg_logger;
                        $skip_line = 1;
                        last;
                    }

                    $rec->{ $field_def->{field_name} } = $rv;
                }
                else {
                    $rec->{ $field_def->{field_name} } = $field_value;
                }
            }
            next if $skip_line;

            my $message_protobuf_class_name = $protobuf_convert_cfg->{message_protobuf_class_name};
            my $message = $message_protobuf_class_name->new($rec);
            my $protobuf_class_enum = $protobuf_convert_cfg->{protobuf_class_enum};
            my $bd_config_element = Bd::Config::BdConfigElement->new({
                config_action => $operation,
                bd_table => Bd::Config::BdTable->$protobuf_class_enum(),
                $protobuf_convert_cfg->{protobuf_entry_name} => $message,
            });
            $current_message_size += length($bd_config_element) + 4;
            #protobuf encoded message about twice size than data
            my $encoded_message_size = 2 * $current_message_size;

            if ($protobuf_message_max_size &&
                scalar(@config_elements) &&
                ($encoded_message_size >= $protobuf_message_max_size || scalar(@config_elements) >= $PROTOBUF_MAX_BULK_SIZE)) {

                $bd_conf_elem_from_prev_iteration = $bd_config_element;
                last;
            }
            else {
                push @config_elements, $bd_config_element;
                $operation_count->{$operation}++;
            }
        }

        $cfg_logger->trace("Completed BD config from $from_file") if $cfg_logger;

        my $records_count = 1;
        $last_iteration = 0;
        if (! scalar(@config_elements)) {
           $records_count = 0;
           $last_iteration = 1;
           close $inf;
           $inf = undef;
        }

        my $bd_config = Bd::Config::BdConfig->new({ conf_element => \@config_elements });
        $cfg_logger->trace("Sending BdConfig protobuf message for $protobuf_convert_cfg->{message_protobuf_class_name}") if $cfg_logger;
        $cfg_logger->trace($bd_config) if $cfg_logger;

        my $serialized_config = $bd_config->encode;
        my $prefix =  pack($PREFIX_PACK_STRING,
                           $PROTOBUF_DCC_TABLE_ID,
                           length($serialized_config),
                           $records_count);

        my %metadata = (
            added_records => $operation_count->{ Bd::Config::ConfigAction::ADD() } || 0,
            updated_records => $operation_count->{ Bd::Config::ConfigAction::UPDATE() } || 0,
            deleted_records => $operation_count->{ Bd::Config::ConfigAction::DELETE() } || 0,
        );

        return [$prefix . $serialized_config, $last_iteration, %metadata];
    };

    if ($protobuf_message_max_size) {
        return wantarray ? ($iterator) : $iterator ;
    }
    else {
        my ($config_data, $last_iteration, %metadata) = @{$iterator->()};
        return wantarray ? ($config_data, %metadata) : $config_data ;
    }

}

#----------------------------------------------------------------------

sub binary_convert {
    my(%params) = @_;

    my $table_def = $params{table_def};
    my $from_file = $params{from_file};
    my $to_file = $params{to_file};
    my $return_as_string = $params{return_as_string};
    my $cfg_logger = $params{cfg_logger};
    my $retval = '';
    my $outf;

    my $operation_count;

    $cfg_logger->trace("Converting BD config from $from_file") if $cfg_logger;
    open(my $inf, '<', $from_file);
    if ($to_file) {
        open($outf, '>', $to_file);
        syswrite($outf, pack('L a36', $MSG_SIGNATURE_DEFAULT, ''));
        syswrite($outf, pack('i i i', 0, 0, 0)); # place holder for block count - only determined by the end
    }

    my $binary_convert_cfg = get_binary_convert_cfg(table_def => $table_def);
    my $pack_string = $binary_convert_cfg->{pack_string};
    my $row_length = $binary_convert_cfg->{row_length};

    my $row_count = 0;
    while (my $line = <$inf>) {
        chomp($line);
        $cfg_logger->trace("Cfg line = $line") if $cfg_logger;
        my(@field_vals) = split(/,/, $line);

        my $operation;
        if ($table_def->{send_all_data_in_update}) {
            $operation = Bd::Config::ConfigAction::ADD();
        }
        else {
            my $rec_operation = $field_vals[0];
            $operation = $PROTOBUF_OPERATION_MAP{ $rec_operation };
            $operation ||= Bd::Config::ConfigAction::ADD();
        }

        my $skip_line = 0;
        my $i = $#field_vals;
        while ($i >= 0) {
            my $field_def = $binary_convert_cfg->{field_def}->[$i];

            if ($field_def->{fixup_func}) {
                my @rv;
                eval {
                    @rv = $field_def->{fixup_func}->($field_vals[$i]);
                };
                if (my $err = $@) {
                    $cfg_logger->error("Error encountered while handling config line = $line, element = $i ($err)") if $cfg_logger;
                    $skip_line = 1;
                    last;
                }

                splice(@field_vals, $i, 1, @rv);
            }

            if ($field_vals[$i] eq '' && exists $field_def->{empty_value}) {
                $field_vals[$i] = $field_def->{empty_value};
            }

            $i--;
        }
        next if $skip_line;
        $operation_count->{$operation}++;

        my $encoded_row = pack($pack_string, @field_vals);
        $row_count++;

        if ($to_file) {
            syswrite($outf, $encoded_row);
        }

        if ($return_as_string) {
            $retval .= $encoded_row;
        }
    }
    close $inf;
    $cfg_logger->trace("Completed BD config from $from_file") if $cfg_logger;

    if ($row_count == 0) {
        # Old cfg_conventer sends row_length = 0 when there are no records
        $row_length = 0;
    }

    my $prefix = pack($PREFIX_PACK_STRING,
                      $table_def->{dcc_table_id},
                      $row_length,
                      $row_count);
    if ($to_file) {
        # Rewrite block size and row count
        sysseek($outf, 40, SEEK_SET);
        syswrite($outf, $prefix, 12);

        close $outf;
    }

    if ($return_as_string) {
        $retval = $prefix . $retval;
    }

    my %metadata = (
        added_records => $operation_count->{ Bd::Config::ConfigAction::ADD() } || 0,
        updated_records => $operation_count->{ Bd::Config::ConfigAction::UPDATE() } || 0,
        deleted_records => $operation_count->{ Bd::Config::ConfigAction::DELETE() } || 0,
    );

    return wantarray ? ($retval, %metadata) : $retval ;
}

#----------------------------------------------------------------------

sub get_binary_convert_cfg {
    my(%params) = @_;

    my $table_def = $params{table_def};

    if (! $table_def->{binary_convert_cfg}) {

        my @field_def;
        my $pack_string;
        my $row_length;
        if (! $table_def->{send_all_data_in_update}) {
            $pack_string = 'c';
            $row_length = 1;
            push(@field_def, $types{'byte'});
        }

        for my $field (@{$table_def->{fields}}) {
            my $data_type = $field->{data_type};
            my $field_length;
            if ($types{$data_type}) {
                $field_length = $types{$data_type}->{field_length};
                $pack_string .= $types{$data_type}->{pack_as} . ' ';
            }
            else {
                $data_type =~ s/(\d+)$//;
                $field_length = $1;
                $pack_string .= $types{$data_type}->{pack_as} . $field_length . ' ';
            }
            $row_length += $field_length;

            push(@field_def, $types{$data_type});
        }

        $table_def->{binary_convert_cfg} = {
                                            field_def => \@field_def,
                                            row_length => $row_length,
                                            pack_string => $pack_string,
                                           };
    }

    return $table_def->{binary_convert_cfg};
}

#----------------------------------------------------------------------

sub get_dcc_table_config {
    if (! $dcc_table_config) {
        $dcc_table_config = YAML::XS::LoadFile($DCC_TABLE_CFG_FILENAME);
        for my $dcc_table_rec (@$dcc_table_config) {
            unless ($dcc_table_rec->{output_file}) {
                $dcc_table_rec->{output_file} = lc $dcc_table_rec->{table};
                $dcc_table_rec->{output_file} =~ s/^dcc\.//;
            }
        }
    }
    return $dcc_table_config;
}

#----------------------------------------------------------------------

1;


