package F5::ObjUtils;

use strict;
use warnings;
use Scalar::Util qw(reftype);
use Storable;
use Exporter ();
our @ISA = qw(Exporter);
our @EXPORT = qw(dclone freeze store_fd);

$Storable::canonical = 1;

my $DEFAULT_RECURSIVE_DEPTH_LIMIT = 10;
my %UNFREEZABLE_TYPES = ( 'DBI::db' => 1,
                          'CODE' => 1,
                          'F5::ASM::COAPI::Object' => 1 );

#----------------------------------------------------------------------

sub is_empty {
    my ($obj) = @_;
    my $reftype = reftype($obj);
    if (!defined($obj) ||
        ($reftype eq 'ARRAY' && !@$obj) ||
        ($reftype eq 'HASH' && !%$obj)
       ) {
        return 1;
    }
    return 0;
}

#----------------------------------------------------------------------

sub dclone {
  my ($obj) = @_;
  remove_unfreezable_references($obj);
  return Storable::dclone($obj);
}

#----------------------------------------------------------------------

sub freeze {
  my ($obj) = @_;
  remove_unfreezable_references($obj);
  return Storable::freeze($obj);
}


#----------------------------------------------------------------------

sub store_fd {
  my ($params, $fh) = @_;
  remove_unfreezable_references($params);
  return Storable::store_fd($params, $fh); 
}

#----------------------------------------------------------------------

sub remove_unfreezable_references {
  my($obj, %params) = @_;

  $params{seen_refs} ||= {};
  $params{depth_count} ||= 0;
  $params{recursive_depth_limit} ||= $DEFAULT_RECURSIVE_DEPTH_LIMIT;
  
  return if ($params{depth_count} >= $params{recursive_depth_limit});
  return if (! $obj || ! ref($obj));
  return if ($params{seen_refs}->{$obj});

  $params{seen_refs}->{$obj} = 1;
  $params{depth_count}++;

  if (reftype($obj) eq 'ARRAY') {
    for my $i (reverse 0..$#$obj) {
      if (my $ref_type = ref($obj->[$i])) {
        if ($UNFREEZABLE_TYPES{$ref_type}) {
          splice(@$obj, $i, 1);
        }
        else {
          remove_unfreezable_references($obj->[$i], %params);
        }
      }
    }
  }
  elsif (reftype($obj) eq 'HASH')  {
    for my $k (keys %$obj) {
      if (my $ref_type = ref($obj->{$k})) {
        if ($UNFREEZABLE_TYPES{$ref_type}) {
          delete $obj->{$k};
        }
        else {
          remove_unfreezable_references($obj->{$k}, %params);
        }
      }
    }
  }
}

#----------------------------------------------------------------------

sub get_deeply {
    my(%params) = @_;
    my $ref = $params{ref};
    my $search_key = $params{search_key};

    if (ref($search_key) eq 'ARRAY') {
        my @search_list = @$search_key;
        my $search_element = shift @search_list;
        my @element_search_results = get_deeply(ref => $ref, search_key => $search_element);
        if (@search_list) {
            return get_deeply(ref => \@element_search_results, search_key => \@search_list);
        }
        else {
            return @element_search_results;
        }
    }

    if (ref($ref)) {
        if (ref($ref) eq 'ARRAY') {
            return map { get_deeply(%params, ref => $_) } @$ref;
        }
        else {
            my @rv;
            if ($ref->{$search_key}) {
                if (ref($ref->{$search_key}) && ref($ref->{$search_key}) eq 'ARRAY') {
                    @rv = @{$ref->{$search_key}};
                }
                else {
                    @rv = ($ref->{$search_key});
                }
            }
            return (@rv, map { get_deeply(%params, ref => $ref->{$_}) } grep { $_ ne $search_key } keys %$ref);
        }
    }
    return;
}

#----------------------------------------------------------------------

sub recursively_callback {
    my (%params) = @_;
    my $data_ref = $params{data_ref};
    my $callback = $params{callback};

    return unless ref($data_ref);

    $params{seen_refs} ||= {};
    return if ($params{seen_refs}->{$data_ref});
    $params{seen_refs}->{$data_ref} = 1;

    if ( ref $data_ref && ref $data_ref eq 'HASH' ) {
        foreach my $val ( values %$data_ref ) {
            recursively_callback(%params,
                                 data_ref => ( ref($val) ) ? $val : \$val ,
                                );
        }
    }
    elsif( ref $data_ref && ref $data_ref eq 'ARRAY' ) {
        foreach my $val ( @$data_ref ) {
            recursively_callback(%params,
                                 data_ref => ( ref($val) ) ? $val : \$val ,
                                );
        }
    }
    else {
        $callback->(val_ref => $data_ref);
    }
}

#----------------------------------------------------------------------

1;
