package F5::PrepareFiles;

use strict;
use Cwd;

use F5::Deploy;
use F5::Logger qw(:LEGACY_ERROR_CONSTANTS);

use F5::Cfg;
use F5::SetActive;
use F5::DbUtils;
use F5::Provisioning;
use F5::PrepareConf::Base;
use F5::ASM::COAPI;
use F5::PrepareConf::BotDefense;
use F5::Utils::File;

use File::Copy;
use File::Basename;
use File::Path;
use File::Find;
use YAML::XS;
use MIME::Base64;
use Time::HiRes;
use POSIX qw/strftime/;

use F5::BigipVersionUtils;

use vars qw{$RET_OK
	    $RET_NOT_OK
	    $LOG_FILE
	    $PRINT_TARGET
	    $DEBUG
	    };

#default log file for this module, may be changed from upper level
$LOG_FILE = 'prepare_files.log';
$PRINT_TARGET = 3;
my $CONF_FILE = "/etc/ts/dcc/prepare_files.cfg";
my $CHECKSUM_FILE = F5::Deploy::map_path('/ts/recovery/internal_config_checksum');

my $DCC_TABLE_CFG;
my %DCC_TABLE_CFG;
my $DCC_TABLE_CFG_FILENAME = F5::Deploy::map_path('/etc/ts/common/dcc_tables_cfg.yaml');
my $LOGGER;

my $ROW_ACTION_KEEP = 1;
my $ROW_ACTION_GET_NEXT = 2;
my $ROW_ACTION_FINISHED = 3;
$DEBUG = 0;

$RET_OK = 0;
$RET_NOT_OK = -1;

#--------------------------------------------------------------------------------
sub PrepareSiteTree
{
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $test_site_dir = $params{test_site_dir};

    my $site_dir = (defined $test_site_dir) ? $test_site_dir. '/files/site_1' : '/ts/files/site_' . $site_id;
    pf_logprint($INF,undef,"started, site_pid=$site_id\n");

    my $conf_dir = F5::Deploy::map_path($site_dir . '/config');

    F5::Utils::File::delete_files_recursively($conf_dir) if(-d $conf_dir);

    my @SITE_CONFIG_DIRS = qw(full full/package fresh update);
    for my $dir (@SITE_CONFIG_DIRS) {
        File::Path::make_path("$conf_dir/$dir");
    }

    return $RET_OK;
}

#--------------------------------------------------------------------------------
sub PrepareInternalConfigFiles
{
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $test_site_dir = $params{test_site_dir};
    my $dbh = $params{dbh};

    my $dest_dir = (defined $test_site_dir) ? $test_site_dir : F5::Deploy::map_path('/ts/files/site_' . $site_id.'/config/full');
    pf_logprint($INF,undef,"started, site_id=$site_id\n");

    if ( !$dbh)
    {
	pf_logprint($ERR, 'PrepareInternalConfigFiles', "Missing dbh");
	return $RET_NOT_OK;
    }

    if (!F5::BigipUtils::is_docker()) {
        if (open (my $fh, '>', $CHECKSUM_FILE)) {
            my ($junk, $current_checksum) = $dbh->selectrow_array("CHECKSUM TABLE DCC.INTERNAL_CONFIG_ITEMS");
            print $fh $current_checksum;
            close ($fh);
        }
        else {
            pf_logprint($ERR, 'PrepareInternalConfigFiles', "Could not open $CHECKSUM_FILE");
            return $RET_NOT_OK;
        }
    }

    my $sql = <<EndQuery; 
SELECT cfg_files.file_name,
       cfg_sections.section_name,
       cfg_items.item_id,
       cfg_items.field_name,
       cfg_items.current_value,
       cfg_items.internal_remark
 FROM DCC.INTERNAL_CONFIG_FILES cfg_files
   JOIN DCC.INTERNAL_CONFIG_SECTIONS cfg_sections USING (file_id)
   JOIN DCC.INTERNAL_CONFIG_ITEMS cfg_items USING (section_id)
 WHERE cfg_files.file_name IS NOT NULL 
 ORDER BY cfg_sections.section_order, cfg_items.item_order, cfg_items.item_id
EndQuery

    # read all config items

    my $sth = $dbh->prepare($sql);

    if (not $sth)
    {
	pf_logprint($ERR, 'DBI->prepare',"'$sql': '$DBI::errstr'");
	return $RET_NOT_OK;
    }

    if (not $sth->execute())
    {
	pf_logprint($ERR, 'DBI->execute',"'$sql': '$DBI::errstr'");
	return $RET_NOT_OK;
    }

    # get all rows

    my %config_hash = ();

    # keep track of order of sections
    my $order = 1;

    while(my $row = $sth->fetchrow_hashref)
    {
        my ($file, $section, $item_id, $key, $value, $comment) =
            @$row{qw(file_name section_name item_id field_name current_value internal_remark)};

        # avoid using undef
        $section = '' if not defined $section;

        if (! defined $file || $file eq '')
        {
            pf_logprint($INF, undef, "No file_name for item_id='$item_id', skipping");
            next;
        }

        $config_hash{$file}{$section} ||= { order => $order, items => [] };

        # make sure value has no CRLF
        $value =~ s/\r?\n/ /g;

        # check for comment before item itself
        if (defined $comment && $comment ne '')
        {
            # make sure comment has no CRLF
            $comment =~ s/\r?\n/ /g;
            push @{ $config_hash{$file}{$section}{items} }, "// $comment\n";
        }

        push @{ $config_hash{$file}{$section}{items} }, "$key=$value\n";

        $order++;
    }

    $sth->finish;

    # write out cfg files

    my $ret = $RET_OK;

    foreach my $filename (keys %config_hash)
    {
        my $output_file = "$dest_dir/$filename.cfg";

        pf_logprint($DBG, undef, "writing config file $output_file");

        my $filehash = $config_hash{$filename};

        if (not open OUTPUT, '>', $output_file)
        {
            pf_logprint($ERR, undef, "Cannot open cfg file $output_file ($!)");
            $ret = $RET_NOT_OK;
            next;
        }

        # loop through sections in order
        foreach my $section (sort { $filehash->{$a}{order} <=> $filehash->{$b}{order} } keys %$filehash)
        {
            # pf_logprint($DBG, undef, "writing section '$section'");

            # only write section if there is one defined
            if ($section ne '')
            {
                print OUTPUT "\n[$section]\n\n";
            }

            # loop through each item
            foreach (@{ $filehash->{$section}{items} })
            {
                print OUTPUT;
            }
        }

        close OUTPUT;
    }

    pf_logprint($INF,undef,"ended, site_id=$site_id\n");

    return $ret;
}

#--------------------------------------------------------------------------------
sub PrepareCoreFiles
{
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $dbh = $params{dbh};
    my $site_dir = (defined $params{test_site_dir}) ? $params{test_site_dir} 
                                                    : "/ts/files/site_$site_id";
    my $saved_umask = umask(0000);

    my $conf_dir = F5::Deploy::map_path("$site_dir/config");
    pf_logprint($INF,undef,"started, site_pid=$site_id\n");   
    #clean full dir
  ###  rm("-rf", $conf_dir.'/full/*.cfg');

    set_bigip_info($dbh);

    # Signature names may be affected by changes in INTERNAL_CONFIG_ITEMS, and need to be
    # prepared in case remote_logger_include_sig_risk_accuracy has changed in the interim.
    my $prepare_conf = F5::PrepareConf::Base->new(dbh => $dbh);
    $prepare_conf->prepare_negsig_enforcer_names;
    $prepare_conf->prepare_server_technologies_fingerprints;
    $prepare_conf->prepare_icap_server();
    $prepare_conf->prepare_database_protection_server();
    $prepare_conf->prepare_viol_rating;
    $prepare_conf->prepare_enforcer_accounts();
    $prepare_conf->persist_to_mcp();
    $prepare_conf->prepare_data_protection();
    $prepare_conf->prepare_deception_response_pages();
    $prepare_conf->prepare_deception_traps();

    my $prepare_core_files_result = prepare_core_files(dbh => $dbh,
                                                       dest_dir => $conf_dir,
                                                       prepare_account_tables => 1,
                                                      );

    my $bot_defense = F5::PrepareConf::BotDefense->new(dbh => $dbh,
                                                       logger => $LOGGER,
                                                      );
    $bot_defense->prepare_bot_defense_profiles();

    pf_logprint($INF,undef,"ended, site_pid=$site_id\n");
    umask($saved_umask);
    return ($prepare_core_files_result->{completed_successfully} ? $RET_OK : $RET_NOT_OK);
}
#--------------------------------------------------------------------------------
sub set_bigip_info {
    my ($dbh) = @_;

    F5::DbUtils::update_row(dbh => $dbh,
                            table_name => 'DCC.BIG_IP_UNIT_INFO',
                            rec => {
                                    software_version => F5::BigipVersionUtils::bigip_version(),
                                    flg_is_asm_provisioned => F5::Provisioning::is_module_enabled('asm'),
                                   },
                           );
}

#--------------------------------------------------------------------------------
sub PrepareCoreFilesUpdate {
    my (%params) = @_;

    my $site_id = $params{site_id};
    my $dbh = $params{dbh};
    my $site_dir = (defined $params{test_site_dir}) ? $params{test_site_dir} 
                                                    : "/ts/files/site_$site_id";
    my $conf_dir = F5::Deploy::map_path("$site_dir/config");

    my $logger = $LOGGER;
    my $policy = $params{policy};
    my $original_policy_name_crc = $params{original_policy_name_crc};

    my ($time, $micro) = Time::HiRes::gettimeofday();
    my $run_time = POSIX::strftime("%Y%m%d%H%M%S", localtime($time)) . sprintf('.%03d', int($micro / 1000));
    $logger->info("Started (run_time = $run_time)") if $logger;
    my $update_dir = "$conf_dir/update/$run_time";
    if (! -e $update_dir) {
        mkpath($update_dir);
    }

    my $prepare_core_files_result = prepare_core_files(dbh => $dbh,
                                                       dest_dir => $conf_dir,
                                                       policy => $policy,
                                                       generate_update => 1,
                                                       update_run_time => $run_time,
                                                       update_dir => $update_dir,
                                                       original_policy_name_crc => $original_policy_name_crc,
                                                       prepare_account_tables => $params{prepare_account_tables},
                                                       modified_dcc_tables => $params{modified_dcc_tables},
                                                      );

# Uncomment for debugging only
###    save_conf_update_files_for_debug($site_dir);
    $logger->info("Finished (run_time = $run_time)") if $logger;

    return { completed_successfully => $prepare_core_files_result->{completed_successfully},
             update_dir => $update_dir,
           };
}

#--------------------------------------------------------------------------------

sub cleanup_update_files {
    my (%params) = @_;

    my $site_id = $params{site_id} || 1;
    my $site_dir = (defined $params{test_site_dir}) ? $params{test_site_dir} : "/ts/files/site_$site_id";
    my $update_base_dir = F5::Deploy::map_path("$site_dir/config/update");

    F5::Utils::File::cleanup_files(logger => $LOGGER,
                                   dir => $update_base_dir,
                                   max_size => 5 * 1024 * 1024, # 5 MB
                                  );
}

#----------------------------------------------------------------------
# prepare_core_files builds the full DCC config files, either for all accounts or
# for a particular account (if specified). Wherever possible, config is separated
# into distinct files for each account.
#
# Additionally, prepare_core_files builds update DCC config difference files,
# relative to the last configuration. The update files are composed by iterating
# simultaneously through the DCC database cursor, aka the "fresh" row, and the
# last full config file, aka the "full" row. The current fresh row and current
# full row are compared lexically to determine:
#
#  - If the fresh row keys precede the full row keys, this is a new fresh row and
#    it is added to the update as "+1" operation.
#
#  - If the full row keys precede the fresh row keys, this is a deleted full row
#    and it is added to the update as "-1" operation.
#
#  - If the fresh and full rows have matching keys, but differ in value, this is
#    an updated row and it is added to the update as "0" operation.
#
sub prepare_core_files {
    my(%params) = @_;

    my $dbh = $params{dbh};
    my $dest_dir = $params{dest_dir};
    my $logger = $LOGGER;
    my $generate_update = $params{generate_update} || 0;
    my $policy = $params{policy};
    my $original_policy_name_crc = $params{original_policy_name_crc};
    my $prepare_account_tables = $params{prepare_account_tables};
    my %modified_dcc_tables;
    if ($params{modified_dcc_tables}) {
        %modified_dcc_tables = map { ($_ => 1) } @{$params{modified_dcc_tables}};
    }
    my $update_dir = $params{update_dir};

    prepare_violations_for_mcp(%params);
    prepare_attack_types_for_mcp(%params);

    my $policy_name_crc = 0;
    if ($original_policy_name_crc) {
        $policy_name_crc = $original_policy_name_crc;
    }
    elsif ($policy && $policy->{policy_name_crc}) {
        $policy_name_crc = $policy->{policy_name_crc};
    }

    my @generated_dcc_tables;

    for my $dcc_table (@$DCC_TABLE_CFG) {
        if ($dcc_table->{split_config_by_account} && ! $prepare_account_tables) {
            # Skip account-specific table for non-account update (such as for logging profile)
            next;
        }

        if (%modified_dcc_tables && ! $modified_dcc_tables{$dcc_table->{table}}) {
            $logger->trace("Skipping $dcc_table->{table} (not specified in table list)") if $logger;
            next;
        }

        push(@generated_dcc_tables, $dcc_table);

        my ($cfg_query, @bind_params, @field_types);
        my $full_file = {};
        my ($full_row, $fresh_row);

        my $update_add_file = { dest_dir => $update_dir,
                                prefix =>  sprintf('%s__ADD__%04d__', $params{update_run_time}, $#generated_dcc_tables + 1),
                                dcc_table => $dcc_table,
                               };
        my $update_del_file = $update_add_file;
        if ($dcc_table->{split_adds_and_deletes}) {
            # Adds and deletes for a particular DCC table may need to be split in order to allow foreign keys to
            # switch over to the new record before the old one is deleted (such as XML profiles, which are referenced
            # from URLs). For such table, deletes will be handled at the end, after all adds and updates have completed.
            #
            # To ensure that foreign key order is respected, deletes must be handled in reverse order. This is accomplished
            # by subtracting the current update index from 10k (there should be fewer than 5k DCC tables for a long long time).
            $update_del_file = { %$update_add_file,
                                 prefix =>  sprintf('%s__DEL__%04d__', $params{update_run_time}, 10_000 - $#generated_dcc_tables),
                               };
        }

        my $fresh_file = { dest_dir => "$dest_dir/fresh", 
                           dcc_table => $dcc_table,
                         };

        my @full_files;
        if ($generate_update) {
            if ($dcc_table->{split_config_by_account}) {
                if ($policy_name_crc) {
                    @full_files = ("$dest_dir/full/$dcc_table->{output_file}__account_$policy_name_crc.cfg");
                }
                else {
                    @full_files = (sort glob("$dest_dir/full/$dcc_table->{output_file}__account*.cfg"));
                }
            }
            else {
                @full_files = ("$dest_dir/full/$dcc_table->{output_file}.cfg");
            }
        }

        if ($dcc_table->{legacy_query}) {
            $cfg_query = $dcc_table->{legacy_query};
        }
        else {
            my @field_names = map { $_->{field_name} } @{$dcc_table->{fields}};
            @field_types = map { $_->{data_type} } @{$dcc_table->{fields}};

            if ($dcc_table->{split_config_by_account} && ! $dcc_table->{fields_by_name}->{policy_name_crc}) {
                unshift(@field_names, 'policy_name_crc');
            }
            my $field_names = join(', ', @field_names);

            my @where_clauses;
            if ($policy_name_crc && $dcc_table->{split_config_by_account}) {
                push(@where_clauses, 'policy_name_crc = ?');
                push(@bind_params, $policy_name_crc);
            }
            my $where_clauses;
            if (@where_clauses) {
                $where_clauses = 'WHERE ' . join(' AND ', @where_clauses);
            }

            my $order_clause = '';
            if ($dcc_table->{order_by_fields} && @{$dcc_table->{order_by_fields}}) {
                if (F5::DbUtils::is_sqlite($dbh)) {
                    $order_clause .= " ORDER BY " . join(', ', map { "CAST($_ AS TEXT)" } @{$dcc_table->{order_by_fields}});
                }
                else {
                    $order_clause .= " ORDER BY " . join(', ', map { "BINARY $_ " } @{$dcc_table->{order_by_fields}});
                }
            }

            $cfg_query = <<EndQuery;
SELECT $field_names
  FROM $dcc_table->{table}
$where_clauses
$order_clause
EndQuery
        }

        $logger->trace("About to execute $cfg_query") if $logger;
        my $values_sth = $dbh->prepare($cfg_query);
        $values_sth->execute(@bind_params);

        my $fresh_row_action = $ROW_ACTION_GET_NEXT;
        my $full_row_action = $ROW_ACTION_GET_NEXT;
        if (! $generate_update || $dcc_table->{send_all_data_in_update} || ! @full_files) {
            $full_row_action = $ROW_ACTION_FINISHED;
        }

        while ($fresh_row_action != $ROW_ACTION_FINISHED || $full_row_action != $ROW_ACTION_FINISHED) {

            if ($fresh_row_action == $ROW_ACTION_GET_NEXT) {
                # Read next row from db cursor
                my @row = $values_sth->fetchrow_array();
                if (! @row) {
                    $fresh_row = {};
                    $fresh_row_action = $ROW_ACTION_FINISHED;
                }
                else {
                    my @row_compare_policy_name_crc;
                    if ($dcc_table->{split_config_by_account} && ! $dcc_table->{fields_by_name}->{policy_name_crc}) {
                        @row_compare_policy_name_crc = (shift(@row));
                    }

                    if ($dcc_table->{has_transformable_fields}) {

                        for my $char_field_index (@{$dcc_table->{char_field_indexes}}) {
                            if (defined($row[$char_field_index])) {
                                $row[$char_field_index] =~ tr/,\r\n/\002\003\004/;
                            }
                        }
                        for my $binary_field_index (@{$dcc_table->{binary_field_indexes}}) {
                            if (defined($row[$binary_field_index])) {
                                $row[$binary_field_index] = MIME::Base64::encode_base64($row[$binary_field_index], '');
                            }
                        }
                        for my $nullable_field_index (@{$dcc_table->{nullable_field_indexes}}) {
                            if (! defined($row[$nullable_field_index])) {
                                $row[$nullable_field_index] = '';
                            }
                        }
                    }

                    # Compose cfg line from db
                    $fresh_row = { compare_fields => [ @row_compare_policy_name_crc, @row[@{$dcc_table->{order_by_field_indexes}}] ] };

                    $fresh_row->{cfg_line} = join(',', @row) . "\n";
                    if (! $dcc_table->{send_all_data_in_update}) {
                        $fresh_row->{cfg_line} = "+1,$fresh_row->{cfg_line}";
                    }

                    if ($dcc_table->{split_config_by_account}) {
                        if ($dcc_table->{fields_by_name}->{policy_name_crc}) {
                            $fresh_row->{policy_name_crc} = $row[$dcc_table->{fields_by_name}->{policy_name_crc}->{field_index}];
                        }
                        else {
                            $fresh_row->{policy_name_crc} = $row_compare_policy_name_crc[0];
                        }
                    }

                    write_cfg_line(output_file => $fresh_file,
                                   row => $fresh_row);

                    $fresh_row_action = $ROW_ACTION_KEEP;
                }
            }

            if ($full_row_action == $ROW_ACTION_GET_NEXT) {
                while ($full_row_action == $ROW_ACTION_GET_NEXT) {
                    if ($full_file->{fh}) {
                        $full_row->{cfg_line} = readline($full_file->{fh});

                        if ($full_row->{cfg_line}) {
                            last;
                        }
                        else {
                            close($full_file->{fh});
                            $full_file->{fh} = undef;
                            $full_row = {};
                        }
                    }

                    if (! @full_files) {
                        $full_row_action = $ROW_ACTION_FINISHED;
                        last;
                    }
                    else {
                        $full_file->{filename} = shift(@full_files);
                        if (! -e $full_file->{filename}) {
                            next;
                        }
                        if ($full_file->{filename} =~ /account_(\d+)/) {
                            $full_file->{policy_name_crc} = $1;
                        }
                        open($full_file->{fh}, '<', $full_file->{filename});
                    }
                }

                if ($full_row->{cfg_line}) {
                    delete $full_row->{compare_fields};
                    if ($dcc_table->{split_config_by_account}) {
                        $full_row->{policy_name_crc} = $full_file->{policy_name_crc};
                    }

                    $full_row_action = $ROW_ACTION_KEEP;
                }
            }

            if ($full_row_action == $ROW_ACTION_FINISHED && $fresh_row_action == $ROW_ACTION_FINISHED) {
                last;
            }

            if ($generate_update) {
                if ($dcc_table->{send_all_data_in_update}) {
                    if ($fresh_row_action != $ROW_ACTION_FINISHED) {
                        write_cfg_line(output_file => $update_add_file,
                                       row => $fresh_row);
                        $fresh_row_action = $ROW_ACTION_GET_NEXT;
                    }
                }
                elsif ($fresh_row->{cfg_line} && $full_row->{cfg_line} &&
                       $fresh_row->{cfg_line} eq $full_row->{cfg_line}) {

                    $full_row_action = $ROW_ACTION_GET_NEXT;
                    $fresh_row_action = $ROW_ACTION_GET_NEXT;
                }
                else {
                    if ($full_row->{cfg_line} && ! $full_row->{compare_fields}) {
                        my @full_fields = split(/,/, $full_row->{cfg_line});
                        chomp($full_fields[-1]);
                        shift(@full_fields); # remove add/delete/update row marker

                        my @row_compare_policy_name_crc;
                        if ($dcc_table->{split_config_by_account} && ! $dcc_table->{fields_by_name}->{policy_name_crc}) {
                            @row_compare_policy_name_crc = ($full_file->{policy_name_crc});
                        }

                        $full_row->{compare_fields} = [ @row_compare_policy_name_crc, @full_fields[@{$dcc_table->{order_by_field_indexes}}] ];
                    }

                    my $field_comparison = compare_fields($fresh_row, $full_row);

                    if (! $fresh_row->{cfg_line} || ($full_row->{compare_fields} && $field_comparison > 0)) {
                        # Add delete record to update
                        write_cfg_line(output_file => $update_del_file,
                                       row => $full_row,
                                       cfg_line_change => '-1');
                        $full_row_action = $ROW_ACTION_GET_NEXT;
                    }
                    elsif (! $full_row->{cfg_line} || ($fresh_row->{compare_fields} && $field_comparison < 0)) {
                        # Newly added record
                        write_cfg_line(output_file => $update_add_file,
                                       row => $fresh_row);
                        $fresh_row_action = $ROW_ACTION_GET_NEXT;
                    }
                    else {
                        if ($full_row->{cfg_line} ne $fresh_row->{cfg_line}) {
                            # Update values for existing record
                            # Updates are handled in the adds file, even for split add/delete tables
                            write_cfg_line(output_file => $update_add_file,
                                           row => $fresh_row,
                                           cfg_line_change => '0');
                        }
                        $full_row_action = $ROW_ACTION_GET_NEXT;
                        $fresh_row_action = $ROW_ACTION_GET_NEXT;
                    }
                }
            }
            else {
                $fresh_row_action = $ROW_ACTION_GET_NEXT;
            }
        }

        if ($update_add_file->{fh}) {
            close($update_add_file->{fh});
        }
        if ($update_del_file->{fh}) {
            close($update_del_file->{fh});
        }
        if ($fresh_file->{fh}) {
            close($fresh_file->{fh});
        }
        $values_sth->finish;
    }

    move_fresh_to_full(%params,
                       generated_dcc_tables => { map { ($_->{output_file} => 1) } @generated_dcc_tables },
                       logger => $logger,
                      );

    return { completed_successfully => 1 };
}

#--------------------------------------------------------------------------------

sub compare_fields {
    my($left_row, $right_row) = @_;

    if (! $left_row->{compare_fields} || ! @{$left_row->{compare_fields}}) {
        if (! $right_row->{compare_fields} || ! @{$right_row->{compare_fields}}) {
            return 0;
        }
        else {
            return 1;
        }
    }
    elsif (! $right_row->{compare_fields} || ! @{$right_row->{compare_fields}}) {
        return -1;
    }

    for my $i (0..$#{$left_row->{compare_fields}}) {
        if (my $cmp_result = $left_row->{compare_fields}->[$i] cmp $right_row->{compare_fields}->[$i]) {
            return $cmp_result;
        }
    }
    return 0;
}

#--------------------------------------------------------------------------------

sub write_cfg_line {
    my(%params) = @_;

    my $output_file = $params{output_file};
    my $row = $params{row};

    my $new_filename;
    if ($row->{policy_name_crc} && $output_file->{dcc_table}->{split_config_by_account}) {
        if (! $output_file->{policy_name_crc} || $row->{policy_name_crc} ne $output_file->{policy_name_crc}) {
            $new_filename = "$output_file->{dest_dir}/$output_file->{prefix}$output_file->{dcc_table}->{output_file}__account_$row->{policy_name_crc}.cfg";
            $output_file->{policy_name_crc} = $row->{policy_name_crc};
        }
    }
    elsif (! $output_file->{fh}) {
        $new_filename = "$output_file->{dest_dir}/$output_file->{prefix}$output_file->{dcc_table}->{output_file}.cfg";
    }

    if ($new_filename) {
        if ($output_file->{fh}) {
            close($output_file->{fh});
        }
        open($output_file->{fh}, '>', $new_filename);
    }

    my $cfg_line = $row->{cfg_line};
    if (defined $params{cfg_line_change}) {
        $cfg_line =~ s/^\+1,/$params{cfg_line_change},/;
    }

    print { $output_file->{fh} } $cfg_line;
}

#--------------------------------------------------------------------------------

sub move_fresh_to_full {
    my(%params) = @_;

    my $dest_dir = $params{dest_dir};
    my $policy = $params{policy};
    my $generated_dcc_tables = $params{generated_dcc_tables};
    my $logger = $params{logger};
    my $original_policy_name_crc = $params{original_policy_name_crc};

    my $policy_name_crc;
    if ($original_policy_name_crc) {
        $policy_name_crc = $original_policy_name_crc;
    }
    elsif ($policy && $policy->{policy_name}) {
        $policy_name_crc = F5::CRC::get_crc($policy->{policy_name});
    }

    my %fresh_files;
    my %fresh_files_basenames;
    find({wanted => sub {
              my $filename = $File::Find::name;
              return unless (-f $filename);

              $fresh_files{$filename} = 1;
              $fresh_files_basenames{basename($filename)} = 1;
          },
          no_chdir => 1,
         },
         "$dest_dir/fresh");

    find({wanted => sub {
              my $filename = $File::Find::name;
              return unless (-f $filename);

              # Ignore full files that are account-specific, but not from the current account
              if ($policy_name_crc) {
                  if ($filename =~ /__account_(\d+)/) {
                      my $file_policy_name_crc = $1;
                      if ($file_policy_name_crc != $policy_name_crc) {
                          return;
                      }
                  }
              }
              my $base_filename = basename($filename);
              my $base_filename_without_account = $base_filename;
              $base_filename_without_account =~ s/(__account_\d+)?\.cfg$//;

              if ($generated_dcc_tables->{$base_filename_without_account} && ! $fresh_files_basenames{$base_filename}) {
                  $logger->debug("Empty generated file deleted - $filename") if $logger;
                  unlink($filename);
              }
          },
          no_chdir => 1,
         },
         "$dest_dir/full");

    for my $updated_fresh_file (keys %fresh_files) {
        my $full_file = $updated_fresh_file;
        $full_file =~ s|^$dest_dir/fresh/|$dest_dir/full/|;
        move($updated_fresh_file, $full_file);
    }
}

#--------------------------------------------------------------------------------
sub pf_logprint
{      
    $LOGGER->log_print(shift, shift, shift);
}

#----------------------------------------------------------------------

sub init {
    my(%params) = @_;
    my $dbh = $params{dbh};

    $PRINT_TARGET = F5::Cfg::cfg_get_config_item($CONF_FILE, "DEBUG",'PrintTarget');
    $LOG_FILE = '/ts/log/' . F5::Cfg::cfg_get_config_item($CONF_FILE, "DEBUG",'LogFile');
    $LOGGER = $params{logger} || F5::Logger->new($LOG_FILE, $PRINT_TARGET);

    $DEBUG = F5::Cfg::cfg_get_config_item($CONF_FILE, "DEBUG",'DEBUG');

    unless ($DCC_TABLE_CFG) {
        $DCC_TABLE_CFG = YAML::XS::LoadFile($DCC_TABLE_CFG_FILENAME);
        for my $dcc_table (@$DCC_TABLE_CFG) {
            unless ($dcc_table->{output_file}) {
                $dcc_table->{output_file} = lc $dcc_table->{table};
                $dcc_table->{output_file} =~ s/^dcc\.//;
            }
            $dcc_table->{char_field_indexes} = [];
            $dcc_table->{binary_field_indexes} = [];
            $dcc_table->{nullable_field_indexes} = [];
            my $dcc_table_db_fields;
            if ($dcc_table->{table}) {
                $dcc_table_db_fields = F5::DbUtils::get_db_fields($dbh, $dcc_table->{table}, undef, 1);
            }
            my $field_index = 0;
            for my $field (@{$dcc_table->{fields}}) {
                $field->{field_index} = $field_index++;
                $dcc_table->{fields_by_name}->{$field->{field_name}} = $field;

                if (index($field->{data_type}, 'char') == 0) {
                    push(@{$dcc_table->{char_field_indexes}}, $field->{field_index});
                }
                if (index($field->{data_type}, 'binary') == 0) {
                    push(@{$dcc_table->{binary_field_indexes}}, $field->{field_index});
                }
                if ($dcc_table_db_fields->{$field->{field_name}} &&
                    $dcc_table_db_fields->{$field->{field_name}}->{nullable}) {

                    push(@{$dcc_table->{nullable_field_indexes}}, $field->{field_index});
                }
            }

            if (@{$dcc_table->{char_field_indexes}} ||
                @{$dcc_table->{binary_field_indexes}} ||
                @{$dcc_table->{nullable_field_indexes}}) {

                $dcc_table->{has_transformable_fields} = 1;
            }

            if (! $dcc_table->{send_all_data_in_update}) {
                if (! defined $dcc_table->{split_config_by_account} && $dcc_table->{fields_by_name}->{policy_name_crc}) {
                    $dcc_table->{split_config_by_account} = 1;
                }
                if (! defined $dcc_table->{split_config_by_account} && $dcc_table->{table} && 
                    $dcc_table_db_fields->{policy_name_crc}) {

                    $dcc_table->{split_config_by_account} = 1;
                }
            }

            if ($dcc_table->{order_by_fields}) {
                $dcc_table->{order_by_fields} = [ split(/,\s*/, $dcc_table->{order_by_fields}) ];
            }
            else {
                $dcc_table->{order_by_fields} = [ map { $_->{field_name} } grep { $_->{is_primary_key} } @{$dcc_table->{fields}} ];

                if (! $dcc_table->{split_config_by_account} && ! $dcc_table->{singleton_row} && ! @{$dcc_table->{order_by_fields}}) {
                    $dcc_table->{order_by_fields} = [ map { $_->{field_name} } @{$dcc_table->{fields}} ];
                }
            }
            $dcc_table->{order_by_field_indexes} = [ map { $dcc_table->{fields_by_name}->{$_}->{field_index} } @{$dcc_table->{order_by_fields}} ];

            # Add policy_name_crc to order by fields after indexes have been set
            if ($dcc_table->{split_config_by_account} && (! @{$dcc_table->{order_by_fields}} || $dcc_table->{order_by_fields}->[0] ne 'policy_name_crc')) {
                unshift(@{$dcc_table->{order_by_fields}}, 'policy_name_crc');
            }

            $DCC_TABLE_CFG{$dcc_table->{output_file}} = $dcc_table;
        }
    }
}

#----------------------------------------------------------------------

# Violations need to be populated in MCP for AVR usage

sub prepare_violations_for_mcp {
    my(%params) = @_;

    my $dbh = $params{dbh};

    my($db_violations) = $dbh->selectall_hashref(<<EndQuery, 'viol_index');
SELECT viol_index, description
  FROM DCC.PARSER_STATUS_CODES
EndQuery

    my $mcp_violations_query = F5::ASM::COAPI::Query->new('asm_violation');
    my %mcp_violations = map { ($_->{code} => $_) } @$mcp_violations_query;

    my %all_viol_indexes = map { ($_ => 1) } (keys %mcp_violations, keys %$db_violations);

    for my $viol_index (keys %all_viol_indexes) {
        my $mcp_violation = $mcp_violations{$viol_index};
        my $db_violation = $db_violations->{$viol_index};

        if ($db_violation) {
            if (! $mcp_violation || $mcp_violation->{name} ne $db_violation->{description}) {
                # Add or update
                if ( $mcp_violation) {
                    $mcp_violation->delete;
                }

                $mcp_violation = F5::ASM::COAPI::Object->new('asm_violation');
                $mcp_violation->{code} = $viol_index;
                $mcp_violation->{name} = $db_violation->{description};
                $mcp_violation->save;
            }
        }
        elsif ($mcp_violation) {
            # No longer exists in MySQL -- needs to be deleted from MCP
            $mcp_violation->delete;
        }
    }
}

#----------------------------------------------------------------------

sub prepare_attack_types_for_mcp {
    my(%params) = @_;

    my $dbh = $params{dbh};

    my($db_attack_types) = $dbh->selectall_hashref(<<EndQuery, 'attack_type_bit_index');
SELECT attack_type_bit_index, attack_type_name
  FROM DCC.ATTACK_TYPE_CODES
EndQuery

    my $mcp_attack_types_query = F5::ASM::COAPI::Query->new('asm_attack_type');
    my %mcp_attack_type_bit_indexes = map { ($_->{attack_type_bit_index} => $_) } @$mcp_attack_types_query;

    my %all_attack_type_bit_indexes = map { ($_ => 1) } (keys %mcp_attack_type_bit_indexes, keys %$db_attack_types);

    for my $attack_type_bit_index (keys %all_attack_type_bit_indexes) {
        my $mcp_attack_type = $mcp_attack_type_bit_indexes{attack_type_bit_index};
        my $db_attack_type = $db_attack_types->{$attack_type_bit_index};

        if ($db_attack_type) {
            if (! $mcp_attack_type || $mcp_attack_type->{name} ne $db_attack_type->{attack_type_name}) {
                # Add or update
                if ( $mcp_attack_type) {
                    $mcp_attack_type->delete;
                }

                $mcp_attack_type = F5::ASM::COAPI::Object->new('asm_attack_type');
                $mcp_attack_type->{attack_type_bit_index} = $attack_type_bit_index;
                $mcp_attack_type->{name} = $db_attack_type->{attack_type_name};
                $mcp_attack_type->save;
            }
        }
        elsif ($mcp_attack_type) {
            # No longer exists in MySQL -- needs to be deleted from MCP
            $mcp_attack_type->delete;
        }
    }
}

#----------------------------------------------------------------------

# Debugging utility to track full/fresh/update files generated by each run
sub save_conf_update_files_for_debug {
    my($site_dir) = @_;

    find({wanted => sub {
              my $filename = $File::Find::name;
              return unless -f $filename;
              my $new_file = $filename;
              $new_file =~ s|$site_dir|/var/tmp/conf_copy/$$|;
              mkpath(dirname($new_file));
              copy($filename, $new_file);
          },
          no_chdir => 1,
         },
         $site_dir);
}

#----------------------------------------------------------------------

#package return:
1;
